/**********************************************************************\
 © COPYRIGHT 2015 Corporation for National Research Initiatives (CNRI);
                        All rights reserved.

        The HANDLE.NET software is made available subject to the
      Handle.Net Public License Agreement, which may be obtained at
          http://hdl.handle.net/20.1000/103 or hdl:20.1000/103
\**********************************************************************/

package net.handle.hdllib.trust;

import java.security.PublicKey;
import java.util.ArrayList;
import java.util.List;

import net.handle.hdllib.HandleValue;
import net.handle.hdllib.Util;
import net.handle.hdllib.ValueReference;

public class ChainVerifier {
    private static final String TRUST_ROOT_HANDLE = "0.0/0.0";
    private static final int MAX_CHAIN_LENGTH = 50;
    
    private JsonWebSignatureFactory signatureFactory = JsonWebSignatureFactory.getInstance();
    private HandleVerifier handleVerifier = new HandleVerifier();
    
    private List<PublicKey> rootKeys;
    
    public ChainVerifier(List<PublicKey> rootKeys) {
        this.rootKeys = rootKeys;
    }
    
    public ChainVerificationReport verifyValues(String handle, List<HandleValue> values, List<IssuedSignature> issuedSignatures) {
        ChainVerificationReport report = new ChainVerificationReport();
        ValuesSignatureVerificationReport valuesReport = handleVerifier.verifyValues(handle, values, issuedSignatures.get(0).jws, issuedSignatures.get(0).issuerPublicKey);
        report.valuesReport = valuesReport;
        List<IssuedSignatureVerificationReport> reports = checkIssuedSignatures(handle, issuedSignatures);
        report.issuedSignatureVerificationReports = reports;
        JsonWebSignature rootSig = issuedSignatures.get(issuedSignatures.size() - 1).jws;
        HandleClaimsSet rootClaims = handleVerifier.getHandleClaimsSet(rootSig);
        if (rootClaims == null) return report;
        if (isRoot(rootClaims.sub, rootClaims.publicKey)) report.rootIsTrusted = true;
        return report;
    }
    
    public ChainVerificationReport verifyChain(List<IssuedSignature> issuedSignatures) {
        ChainVerificationReport report = new ChainVerificationReport();
        List<IssuedSignatureVerificationReport> reports = checkIssuedSignatures(null, issuedSignatures);
        report.issuedSignatureVerificationReports = reports;
        JsonWebSignature rootSig = issuedSignatures.get(issuedSignatures.size() - 1).jws;
        HandleClaimsSet rootClaims = handleVerifier.getHandleClaimsSet(rootSig);
        if (rootClaims == null) return report;
        if (isRoot(rootClaims.sub, rootClaims.publicKey)) report.rootIsTrusted = true;
        return report;
    }
    
    private List<IssuedSignatureVerificationReport> checkIssuedSignatures(String handle, List<IssuedSignature> issuedSignatures) {
        List<IssuedSignatureVerificationReport> result = new ArrayList<IssuedSignatureVerificationReport>();
        for (IssuedSignature issuedSignature : issuedSignatures) { 
            IssuedSignatureVerificationReport report = handleVerifier.verifyIssuedSignature(handle, issuedSignature);
            result.add(report);
        }
        return result;
    }

    private boolean isRoot(String subject, PublicKey publicKey) {
        if (rootKeys == null) {
            System.err.println("Error missing root keys.");
        }
        return Util.equalsPrefixCI(TRUST_ROOT_HANDLE, ValueReference.fromString(subject).getHandleAsString()) && rootKeys.contains(publicKey);
    }

}
